@@ -37,11 +37,11 @@
 #define CTL_INSCOLS	CSI "%u@"
 #define CTL_INSLINES	CSI "%uL"
 #define CTL_RGB		OSC "4;%d;rgb:%02x/%02x/%02x" ST
+#define CTL_CURSORRGB	OSC "12;rgb:%02x/%02x/%02x" ST
 #define CTL_CURSORPOS	CSI "6n"
 #define FG_COLOR	30
 #define BG_COLOR	40
 #define XTERM_TITLE	OSC "0;%s" ST
-/* was: #define XTERM_TITLE	OSC "2;%s" ESC "\\" */
 #define XTERM_ALTSCREEN	CSI "?47h"
 #define XTERM_REGSCREEN	CSI "?47l"
 #define XTERM_BCKISDEL	CSI "?67l"
@@ -49,24 +49,30 @@
 
 #define BAD_RGBA 0x01010100
 
-
 struct cell {
 	Unicode_t unicode;
 	rgba_t fgrgba, bgrgba;
 };
 
+enum cursor_position_knowledge {
+	NEEDED, SOUGHT, KNOWN, INVALID
+};
+
 struct display {
-	unsigned rows, columns;
-	unsigned cursor_row, cursor_column;
+	int rows, columns;
+	int cursor_row, cursor_column;
+	rgba_t cursor_rgba;
 	Boolean_t size_changed;
 	rgba_t fgrgba, bgrgba;
-	unsigned at_row, at_column;
+	int at_row, at_column;
+	enum cursor_position_knowledge get_initial_cursor_position;
+	int initial_row, initial_column;
 	struct cell *image;
 	struct display *next;
 	Byte_t inbuf[INBUF_SIZE];
 	char outbuf[OUTBUF_SIZE];
 	size_t inbuf_bytes, outbuf_bytes;
-	Boolean_t is_xterm, is_linux;
+	Boolean_t is_xterm, is_linux, is_apple;
 	rgba_t color[MAX_COLORS];
 	unsigned colors;
 	char *title;
@@ -132,13 +138,13 @@ static void outf(struct display *display, const char *msg, ...)
 	outs(display, buf);
 }
 
-static void force_moveto(struct display *display, unsigned row, unsigned column)
+static void force_moveto(struct display *display, int row, int column)
 {
 	outf(display, CTL_GOTO, (display->at_row = row) + 1,
 	     (display->at_column = column) + 1);
 }
 
-static void moveto(struct display *display, unsigned row, unsigned column)
+static void moveto(struct display *display, int row, int column)
 {
 	if (row == display->at_row && column == display->at_column)
 		return;
@@ -191,7 +197,7 @@ static unsigned color_delta(rgba_t rgba1, rgba_t rgba2)
 static rgba_t color_mean(rgba_t rgba1, rgba_t rgba2)
 {
 	rgba_t mean = 0;
-	unsigned j;
+	int j;
 	for (j = 0; j < 32; j += 8) {
 		Byte_t c1 = rgba1 >> j, c2 = rgba2 >> j;
 		mean |= c1 + c2 >> 1 << j;
@@ -202,12 +208,14 @@ static rgba_t color_mean(rgba_t rgba1, rgba_t rgba2)
 static unsigned colormap(struct display *display, rgba_t rgba)
 {
 	static rgba_t basic[] = {
-		0, 0x7f000000, 0x007f0000, 0x7f7f0000,
-		0x00007f00, 0x7f007f00, 0x007f7f00, 0x7f7f7f00,
+		BLACK_RGBA, PALE_RGBA(RED_RGBA), PALE_RGBA(GREEN_RGBA),
+		PALE_RGBA(YELLOW_RGBA), PALE_RGBA(BLUE_RGBA),
+		PALE_RGBA(MAGENTA_RGBA), PALE_RGBA(CYAN_RGBA),
+		PALE_RGBA(WHITE_RGBA),
 	};
 	static rgba_t bright[] = {
-		0, 0xff000000, 0x00ff0000, 0xffff0000,
-		0x0000ff00, 0xff00ff00, 0x00ffff00, 0xffffff00,
+		BLACK_RGBA, RED_RGBA, GREEN_RGBA, YELLOW_RGBA,
+		BLUE_RGBA, MAGENTA_RGBA, CYAN_RGBA, WHITE_RGBA,
 	};
 
 	unsigned idx, best, bestdelta, delta;
@@ -274,12 +282,13 @@ static void foreground_color(struct display *display, rgba_t rgba)
 	}
 }
 
-void display_put(struct display *display, unsigned row, unsigned column,
+void display_put(struct display *display, int row, int column,
 		 Unicode_t unicode, rgba_t fgrgba, rgba_t bgrgba)
 {
 	char buf[8];
 	struct cell *cell;
-	if (row >= display->rows || column >= display->columns)
+	if (row < 0 || row >= display->rows ||
+	    column < 0 || column >= display->columns)
 		return;
 	if (unicode < ' ')
 		unicode = ' ';
@@ -301,74 +310,97 @@ void display_put(struct display *display, unsigned row, unsigned column,
 	cell->fgrgba = fgrgba;
 }
 
-static void clear_image(struct display *display, unsigned row, unsigned rows,
-			unsigned column, unsigned columns,
-			Boolean_t colors_trashed)
+static void color_fill(struct display *display,
+		       int row, int rows, int column, int columns,
+		       rgba_t fgrgba, rgba_t bgrgba)
 {
-	int j;
-	while (rows--) {
-		struct cell *cell = &display->image[row++ * display->columns +
+	int r, c;
+	for (r = 0; r < rows; r++) {
+		struct cell *cell = &display->image[(row+r) * display->columns +
 						    column];
-		for (j = 0; j < columns; j++, cell++) {
-			cell->unicode = ' ';
-			if (colors_trashed) {
-				cell->fgrgba = BAD_RGBA;
-				cell->bgrgba = BAD_RGBA;
-			}
+		for (c = 0; c < columns; c++, cell++) {
+			cell->fgrgba = fgrgba;
+			cell->bgrgba = bgrgba;
 		}
 	}
 }
 
-void display_erase(struct display *display, unsigned row, unsigned column,
-		   unsigned rows, unsigned columns)
+/* After an erase, insert, or delete command, update the state.
+ * These commands affect an Apple Terminal differently than other
+ * terminal emulators with respect to the background colors;
+ * Apple fills with the default background, while a regular Xterm
+ * and the Linux console fill with the most recent color selections.
+ */
+static void space_fill(struct display *display, int row, int rows,
+		       int column, int columns)
+{
+	int r, c;
+	for (r = 0; r < rows; r++) {
+		struct cell *cell = &display->image[(row+r) * display->columns +
+						    column];
+		for (c = 0; c < columns; c++, cell++)
+			cell->unicode = ' ';
+	}
+
+	if (display->is_apple)
+		color_fill(display, row, rows, column, columns,
+			   DEFAULT_FGRGBA, DEFAULT_BGRGBA);
+	else
+		color_fill(display, row, rows, column, columns,
+			   display->fgrgba, display->bgrgba);
+}
+
+void display_erase(struct display *display, int row, int column,
+		   int rows, int columns)
 {
 	int r;
 
-	if (row >= display->rows || column >= display->columns)
+	if (row < 0 || row >= display->rows ||
+	    column < 0 || column >= display->columns)
 		return;
 	if (row + rows > display->rows)
 		rows = display->rows - row;
 	if (column + columns > display->columns)
 		columns = display->columns - column;
-	if (!columns)
+	if (rows <= 0 || columns <= 0)
 		return;
 
 	if (!column &&
 	    columns == display->columns &&
 	    row + rows == display->rows) {
-		if (row) {
+		if (!row)
+			outs(display, CTL_ERASEALL);
+		else {
 			moveto(display, row, column);
 			outs(display, CTL_ERASETOEND);
-		} else
-			outs(display, CTL_ERASEALL);
-		clear_image(display, row, rows, column, columns, FALSE);
+		}
 	} else if (column + columns == display->columns) {
 		for (r = 0; r < rows; r++) {
 			moveto(display, row + r, column);
 			outs(display, CTL_ERASELINE);
 		}
-		clear_image(display, row, rows, column, columns, TRUE);
 	} else {
 		for (r = 0; r < rows; r++) {
 			moveto(display, row + r, column);
 			outf(display, CTL_ERASECOLS, columns);
 		}
-		clear_image(display, row, rows, column, columns, FALSE);
 	}
+	space_fill(display, row, rows, column, columns);
 }
 
-void display_insert_spaces(struct display *display, unsigned row,
-			   unsigned column, unsigned spaces, unsigned columns)
+void display_insert_spaces(struct display *display, int row, int column,
+			   int spaces, int columns)
 {
 	struct cell *cell;
 
-	if (row >= display->rows || column >= display->columns)
+	if (row < 0 || row >= display->rows ||
+	    column < 0 || column >= display->columns)
 		return;
 	if (column + columns > display->columns)
 		columns = display->columns - column;
 	if (spaces > columns)
 		spaces = columns;
-	if (!spaces)
+	if (spaces <= 0)
 		return;
 	if (column + columns != display->columns) {
 		moveto(display, row, column + columns - spaces);
@@ -378,21 +410,22 @@ void display_insert_spaces(struct display *display, unsigned row,
 	outf(display, CTL_INSCOLS, spaces);
 	cell = &display->image[row*display->columns + column];
 	memmove(cell + spaces, cell, (columns - spaces) * sizeof *cell);
-	clear_image(display, row, 1, column, spaces, FALSE);
+	space_fill(display, row, 1, column, spaces);
 }
 
-void display_delete_chars(struct display *display, unsigned row,
-			  unsigned column, unsigned chars, unsigned columns)
+void display_delete_chars(struct display *display, int row, int column,
+			  int chars, int columns)
 {
 	struct cell *cell;
 
-	if (row >= display->rows || column >= display->columns)
+	if (row < 0 || row >= display->rows ||
+	    column < 0 || column >= display->columns)
 		return;
 	if (column + columns > display->columns)
 		columns = display->columns - column;
 	if (chars > columns)
 		chars = columns;
-	if (!chars)
+	if (chars <= 0)
 		return;
 
 	moveto(display, row, column);
@@ -401,14 +434,14 @@ void display_delete_chars(struct display *display, unsigned row,
 	outf(display, CTL_INSCOLS, chars);
 	cell = &display->image[row*display->columns + column];
 	memmove(cell, cell + chars, (columns - chars) * sizeof *cell);
-	clear_image(display, row, 1, column + columns - chars, chars, FALSE);
+	space_fill(display, row, 1, column + columns - chars, chars);
 }
 
-static Boolean_t validate(struct display *display, unsigned row,
-			  unsigned column, unsigned *rows, unsigned *columns,
-			  unsigned *lines)
+static Boolean_t validate(struct display *display, int row, int column,
+			  int *rows, int *columns, int *lines)
 {
-	if (row >= display->rows || column >= display->columns)
+	if (row < 0 || row >= display->rows ||
+	    column < 0 || column >= display->columns)
 		return FALSE;
 	if (row + *rows > display->rows)
 		*rows = display->rows - row;
@@ -416,12 +449,17 @@ static Boolean_t validate(struct display *display, unsigned row,
 		*lines = *rows;
 	if (column + *columns > display->columns)
 		*columns = display->columns - column;
-	return *lines && *columns;
+	return *lines > 0 && *columns > 0;
 }
 
-static void insert_whole_lines(struct display *display, unsigned row,
-			       unsigned lines, unsigned rows)
+void display_insert_lines(struct display *display, int row, int column,
+			  int lines, int rows, int columns)
 {
+	if (column ||
+	    columns != display->columns ||
+	    !validate(display, row, column, &rows, &columns, &lines))
+		return;
+
 	if (row + rows != display->rows) {
 		moveto(display, row + rows - lines, 0);
 		outf(display, CTL_DELLINES, lines);
@@ -431,14 +469,19 @@ static void insert_whole_lines(struct display *display, unsigned row,
 	memmove(&display->image[(row + lines) * display->columns],
 		&display->image[row * display->columns],
 		(rows - lines) * display->columns * sizeof *display->image);
-	clear_image(display, row, lines, 0, display->columns, FALSE);
+	space_fill(display, row, lines, 0, display->columns);
 }
 
-static void delete_whole_lines(struct display *display, unsigned row,
-			       unsigned lines, unsigned rows)
+void display_delete_lines(struct display *display, int row, int column,
+			  int lines, int rows, int columns)
 {
 	struct cell *cell;  /* assigned later for safety from resizing */
 
+	if (column ||
+	    columns != display->columns ||
+	    !validate(display, row, column, &rows, &columns, &lines))
+		return;
+
 	moveto(display, row, 0);
 	outf(display, CTL_DELLINES, lines);
 	if (row + rows != display->rows) {
@@ -448,99 +491,19 @@ static void delete_whole_lines(struct display *display, unsigned row,
 	cell = &display->image[row * display->columns];
 	memmove(cell, cell + lines * display->columns,
 		(rows - lines) * display->columns * sizeof *cell);
-	clear_image(display, row + rows - lines, lines, 0, display->columns, FALSE);
-}
-
-static void shortpause(struct display *display, int millisec)
-{
-	struct timespec ts;
-	flush(display);
-	ts.tv_sec = 0;
-	ts.tv_nsec = millisec * 1000 * 1000;
-	nanosleep(&ts, NULL);
-}
-
-static void whole_lines(struct display *display, unsigned row,
-			unsigned lines, unsigned rows,
-			void (*mover)(struct display *, unsigned, unsigned,
-				      unsigned))
-{
-	unsigned amount;
-	for (; lines; lines -= amount) {
-		amount = 1 + (lines > 2) + (lines > 4) + (lines > 8);
-		mover(display, row, amount = 1 + (lines > 2), rows);
-		shortpause(display, 12 - 4 * amount);
-	}
-}
-
-static void copy_line(struct display *display, unsigned to_row,
-		      unsigned from_row, unsigned column, unsigned columns)
-{
-	if (to_row >= display->rows ||
-	    from_row >= display->rows ||
-	    to_row == from_row ||
-	    column >= display->columns)
-		return;
-	if (columns > display->columns - column)
-		columns = display->columns - column;
-	while (columns--) {
-		struct cell *cell =
-			&display->image[from_row * display->columns +
-					column];
-		display_put(display, to_row, column++,
-			    cell->unicode, cell->fgrgba, cell->bgrgba);
-	}
-}
-
-void display_insert_lines(struct display *display, unsigned row,
-			  unsigned column, unsigned lines,
-			  unsigned rows, unsigned columns)
-{
-	int j;
-	if (!validate(display, row, column, &rows, &columns, &lines))
-		return;
-	if (!column && columns == display->columns)
-		whole_lines(display, row, lines, rows,
-			    insert_whole_lines);
-	else {
-		for (j = rows - lines; j-- > 0; )
-			copy_line(display, row + lines + j, row + j,
-				  column, columns);
-		display_erase(display, row, column, lines, columns);
-	}
-}
-
-void display_delete_lines(struct display *display, unsigned row,
-			  unsigned column, unsigned lines,
-			  unsigned rows, unsigned columns)
-{
-	int j;
-
-	if (!validate(display, row, column, &rows, &columns, &lines))
-		return;
-	if (!column && columns == display->columns)
-		whole_lines(display, row, lines, rows,
-			    delete_whole_lines);
-	else {
-		for (j = 0; j < rows - lines; j++)
-			copy_line(display, row + j, row + lines + j,
-				  column, columns);
-		display_erase(display, row + rows - lines, column,
-			      lines, columns);
-	}
+	space_fill(display, row + rows - lines, lines, 0, display->columns);
 }
 
 static struct cell *resize(struct display *display, struct cell *old,
-			   unsigned old_rows, unsigned old_columns,
-			   unsigned new_rows, unsigned new_columns)
+			   int old_rows, int old_columns,
+			   int new_rows, int new_columns)
 {
-	unsigned cells = new_rows * new_columns;
+	int cells = new_rows * new_columns;
 	size_t bytes = cells * sizeof *old;
 	struct cell *new = allocate(bytes);
-	unsigned min_rows = new_rows < old_rows ? new_rows : old_rows;
-	unsigned min_columns = new_columns < old_columns ?
-				new_columns : old_columns;
-	unsigned j, k;
+	int min_rows = new_rows < old_rows ? new_rows : old_rows;
+	int min_columns = new_columns < old_columns ? new_columns : old_columns;
+	int j, k;
 
 	if (!old) {
 		memset(new, 0, bytes);
@@ -573,8 +536,7 @@ static struct cell *resize(struct display *display, struct cell *old,
 	return new;
 }
 
-static void set_geometry(struct display *display,
-			 unsigned rows, unsigned columns)
+static void set_geometry(struct display *display, int rows, int columns)
 {
 	if (display->image &&
 	    display->rows == rows &&
@@ -589,6 +551,8 @@ static void set_geometry(struct display *display,
 	display->size_changed = TRUE;
 	display_erase(display, 0, rows, 0, columns);
 	display_cursor(display, display->cursor_row, display->cursor_column);
+	if (display->get_initial_cursor_position == KNOWN)
+		display->get_initial_cursor_position = INVALID;
 }
 
 static void geometry(struct display *display)
@@ -625,8 +589,7 @@ static void geometry(struct display *display)
 	set_geometry(display, rows, columns);
 }
 
-void display_get_geometry(struct display *display,
-			  unsigned *rows, unsigned *columns)
+void display_get_geometry(struct display *display, int *rows, int *columns)
 {
 	*rows = display->rows;
 	*columns = display->columns;
@@ -636,16 +599,20 @@ void display_get_geometry(struct display *display,
 void display_reset(struct display *display)
 {
 	RELEASE(display->image);
-	if (display->is_xterm)
+	if (display->is_xterm) {
+		if (display->get_initial_cursor_position == NEEDED) {
+			display->get_initial_cursor_position = SOUGHT;
+			outs(display, CTL_CURSORPOS);
+		}
 		outs(display, XTERM_ALTSCREEN);
-	else
+	} else
 		outs(display, CTL_RESET);
 	outs(display, CTL_UTF8
 		      CTL_RESETMODES
 		      CTL_RESETCOLORS);
 	if (display->is_xterm)
 		outs(display, XTERM_BCKISDEL);
-	else
+	if (display->is_linux)
 		outs(display, CTL_NUMLOCK CTL_CLEARLEDS CTL_NUMLOCKLED);
 	display->colors = 0;
 	display->fgrgba = BAD_RGBA;
@@ -653,6 +620,7 @@ void display_reset(struct display *display)
 	geometry(display);
 	force_moveto(display, 0, 0);
 	display->cursor_row = display->cursor_column = 0;
+	display->cursor_rgba = BAD_RGBA;
 	flush(display);
 }
 
@@ -665,7 +633,7 @@ static void sigwinch(int signo, siginfo_t *info, void *data)
 		old_sigwinch(signo, info, data);
 }
 
-#if !(defined __linux__ || defined __APPLE__)
+#if !(defined __linux__ || defined BSD || defined __APPLE__)
 void cfmakeraw(struct termios *termios)
 {
 	termios->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|IXON);
@@ -703,9 +671,16 @@ struct display *display_init(void)
 
 	display_list = display;
 	display->inbuf_bytes = display->outbuf_bytes = 0;
-	term = getenv("TERM");
-	if (!(display->is_xterm = term && !strncmp(term, "xterm", 5)))
-		display->is_linux = term && !strcmp(term, "linux");
+	if (!(term = getenv("TERM"))) {
+	} else if ((display->is_xterm = !strncmp(term, "xterm", 5))) {
+		if ((term = getenv("COLORTERM")) &&
+		    !strncmp(term, "gnome-", 6)) {
+			/* Gnome terminal? Ignore $TERM_PROGRAM */
+		} else if ((term = getenv("TERM_PROGRAM")))
+			display->is_apple = !strcmp(term, "Apple_Terminal");
+	} else {
+		display->is_linux = !strcmp(term, "linux");
+	}
 	display_reset(display);
 	return display;
 }
@@ -720,9 +695,14 @@ void display_end(struct display *display)
 	display_title(display, NULL);
 	set_color(display, DEFAULT_FGRGBA, FG_COLOR);
 	set_color(display, DEFAULT_BGRGBA, BG_COLOR);
-	if (display->is_xterm)
+	if (display->is_xterm) {
 		outs(display, XTERM_REGSCREEN);
-	else
+		if (display->get_initial_cursor_position == KNOWN)
+			force_moveto(display, display->initial_row,
+				     display->initial_column);
+		else
+			force_moveto(display, display->rows-1, 0);
+	} else
 		outs(display, CTL_RESET CTL_RESETMODES);
 	flush(display);
 
@@ -742,35 +722,54 @@ void display_end(struct display *display)
 	RELEASE(display);
 }
 
-void display_title(struct display *display, const char *title)
+Boolean_t display_title(struct display *display, const char *title)
 {
-	if (display->is_xterm) {
-		if (title && !*title)
-			title = NULL;
-		if (!title && !display->title)
-			return;
-		if (!title)
-			RELEASE(display->title);
-		else if (display->title && !strcmp(title, display->title))
-			return;
-		if (title)
-			display->title = strdup(title);
-		else
-			title = "";
-		outf(display, XTERM_TITLE, title ? title : "");
-	}
+	if (!display->is_xterm)
+		return FALSE;
+	if (title && !*title)
+		title = NULL;
+	if (!title && !display->title)
+		return TRUE;
+	if (!title)
+		RELEASE(display->title);
+	else if (display->title && !strcmp(title, display->title))
+		return TRUE;
+	if (title)
+		display->title = strdup(title);
+	else
+		title = "";
+	outf(display, XTERM_TITLE, title ? title : "");
+	return TRUE;
 }
 
-void display_cursor(struct display *display, unsigned row, unsigned column)
+void display_cursor(struct display *display, int row, int column)
 {
 	if (row >= display->rows)
 		row = display->rows - 1;
+	if (row < 0)
+		row = 0;
 	if (column >= display->columns)
 		column = display->columns - 1;
+	if (column < 0)
+		column = 0;
 	display->cursor_row = row;
 	display->cursor_column = column;
 }
 
+Boolean_t display_cursor_color(struct display *display, rgba_t rgba)
+{
+	if (!display->is_xterm ||
+	    display->is_apple ||
+	    rgba & 0xff /* no alpha */)
+		return FALSE;
+	if (rgba != display->cursor_rgba) {
+		outf(display, CTL_CURSORRGB,
+		     rgba >> 24, rgba >> 16 & 0xff, rgba >> 8 & 0xff);
+		display->cursor_rgba = rgba;
+	}
+	return TRUE;
+}
+
 void display_beep(struct display *display)
 {
 	emit("\a", 1);
@@ -783,7 +782,7 @@ Unicode_t display_getch(struct display *display, Boolean_t block)
 	Unicode_t key;
 	unsigned used, vals, val[16];
 
-#define SET_GEOMETRY FUNCTION_F(99)
+#define GOT_CURSORPOS FUNCTION_F(99)
 
 	if (!display)
 		return ERROR_EOF;
@@ -852,7 +851,7 @@ again:	if (display->size_changed)
 		switch (*p) {
 		case 'R': /* cursor position report from southeast corner */
 			if (vals >= 2)
-				key = SET_GEOMETRY;
+				key = GOT_CURSORPOS;
 			break;
 		case '~':
 			if (!val[0])
@@ -936,8 +935,14 @@ done:	if (!key)
 		key = *(p = display->inbuf);
 	used = ++p - display->inbuf;
 	memmove(display->inbuf, p, display->inbuf_bytes -= used);
-	if (key == SET_GEOMETRY) {
-		set_geometry(display, val[0], val[1]);
+	if (key == GOT_CURSORPOS) {
+		if (display->get_initial_cursor_position == SOUGHT &&
+		    val[0] > 0 && val[1] > 0) {
+			display->get_initial_cursor_position = KNOWN;
+			display->initial_row = val[0] - 1;
+			display->initial_column = val[1] - 1;
+		} else
+			set_geometry(display, val[0], val[1]);
 		goto again;
 	}
 	return key;
