@@ -11,8 +11,6 @@
  */
 
 struct text *text_list;
-unsigned default_tab_stop = 8; /* the only correct value :-) */
-Boolean_t default_no_tabs;
 
 struct view *view_find(const char *name)
 {
@@ -92,12 +90,14 @@ struct view *text_create(const char *path, unsigned flags)
 {
 	struct text *text = allocate0(sizeof *text), *prev, *bp;
 
-	if (default_no_tabs)
-		flags |= TEXT_NO_TABS;
-	text->tabstop = default_tab_stop;
 	text->fd = -1;
 	text->flags = flags;
 	text->path = strdup(path);
+	if (utf8_mode == UTF8_NO)
+		text->flags |= TEXT_NO_UTF8;
+	if (default_no_tabs)
+		text->flags |= TEXT_NO_TABS;
+	text->tabstop = default_tab_stop;
 	keyword_init(text);
 
 	for (prev = NULL, bp = text_list; bp; prev = bp, bp = bp->next)
