@@ -10,13 +10,13 @@
 struct window {
 	struct view *view;
 	locus_t start;
-	unsigned row, column;
-	unsigned rows, columns;
-	unsigned cursor_row, cursor_column;
+	int row, column;
+	int rows, columns;
+	int cursor_row, cursor_column;
 	rgba_t fgrgba, bgrgba;
 	unsigned last_dirties;
+	Boolean_t repaint;
 	position_t last_cursor, last_mark;
-	rgba_t last_fgrgba, last_bgrgba;
 	struct mode *last_mode;
 	struct window *next;
 };
@@ -24,30 +24,46 @@ struct window {
 static struct window *window_list;
 static struct window *active_window;
 static struct display *display;
-static unsigned display_rows, display_columns;
+static int display_rows, display_columns;
+static Boolean_t titles = TRUE;
 
 static void title(struct window *window)
 {
 	char buff[128];
+	struct view *view;
+	position_t cursor;
+
+	if (!titles)
+		return;
 	if (window != active_window)
 		return;
 	if (!window) {
-		display_title(display, NULL);
+		titles = display_title(display, NULL);
 		return;
 	}
-	snprintf(buff, sizeof buff, "%s%s",
-		 window->view->name,
-		 window->view->text->flags & TEXT_CREATED ? "(new)" :
-		 window->view->text->flags & TEXT_RDONLY ? "(read-only)" :
-		 window->view->text->preserved !=
-		    window->view->text->dirties ? "(unsaved)" : "");
-	display_title(display, buff);
+	view = window->view;
+	snprintf(buff, sizeof buff, "%s%s", view->name,
+		 view->text->flags & TEXT_CREATED ? " (new)" :
+		 view->text->flags & TEXT_RDONLY ? " (read-only)" :
+		 view->text->preserved !=
+		    view->text->dirties ? " (unsaved)" : "");
+	cursor = locus_get(view, CURSOR);
+	if (cursor < 65536) {
+		int line = current_line_number(view, cursor);
+		int len = strlen(buff);
+		snprintf(buff + len, sizeof buff - len - 1,
+			 " [%d]", line);
+	}
+	titles = display_title(display, buff);
 }
 
 static struct window *activate(struct window *window)
 {
+	if (active_window)
+		active_window->repaint = TRUE;
 	active_window = window;
 	title(window);
+	window->repaint = TRUE;
 	return window;
 }
 
@@ -113,11 +129,13 @@ static Boolean_t window_expand_next(struct window *window)
 	if (stacked(window, next)) {
 		next->row = window->row;
 		next->rows += window->rows;
+		next->repaint = TRUE;
 		return TRUE;
 	}
 	if (beside(window, next)) {
 		next->column = window->column;
 		next->columns += window->columns;
+		next->repaint = TRUE;
 		return TRUE;
 	}
 	return FALSE;
@@ -138,13 +156,15 @@ void window_destroy(struct window *window)
 		struct window *next = previous->next = window->next;
 		if (stacked(previous, window) &&
 		    (!stacked(window, next) ||
-		     previous->rows <= next->rows))
+		     previous->rows <= next->rows)) {
 			previous->rows += window->rows;
-		else if (beside(previous, window) &&
+			previous->repaint = TRUE;
+		} else if (beside(previous, window) &&
 			 (!beside(window, next) ||
-			  previous->columns <= next->columns))
+			  previous->columns <= next->columns)) {
 			previous->columns += window->columns;
-		else if (!window_expand_next(window))
+			previous->repaint = TRUE;
+		} else if (!window_expand_next(window))
 			window_raise(previous->view);
 	} else if ((window_list = window->next) &&
 		   !window_expand_next(window))
@@ -155,13 +175,15 @@ void window_destroy(struct window *window)
 		window->view->window = NULL;
 	}
 	if (window == active_window) {
+		active_window = NULL;
 		wp = window_list;
 		if (previous) {
 			wp = previous;
 			if (wp->next)
 				wp = wp->next;
 		}
-		activate(wp);
+		if (wp)
+			activate(wp);
 	}
 	RELEASE(window);
 }
@@ -205,17 +227,14 @@ struct window *window_after(struct view *before, struct view *view,
 		window->row = old->row;
 		window->rows = old->rows;
 		window->column = old->column +
-			(old->columns -=
-			 (window->columns =
-			  old->columns >> 1));
+			(old->columns -= (window->columns = old->columns >> 1));
 	} else {
 		window->column = old->column;
 		window->columns = old->columns;
 		window->row = old->row +
-			(old->rows -=
-			 (window->rows =
-			  old->rows >> 1));
+			(old->rows -= (window->rows = old->rows >> 1));
 	}
+	old->repaint = TRUE;
 	return activate(window);
 }
 
@@ -263,22 +282,30 @@ struct view *window_current_view(void)
 	return active_window->view;
 }
 
-void windows_end(void)
+void windows_end_display(void)
 {
-	while (window_list)
-		window_destroy(window_list);
+	struct window *window;
 	if (display) {
 		display_end(display);
 		display = NULL;
 	}
+	for (window = window_list; window; window = window->next)
+		window->repaint = TRUE;
+}
+
+void windows_end(void)
+{
+	while (window_list)
+		window_destroy(window_list);
+	windows_end_display();
 }
 
 static unsigned count_rows(struct window *window, position_t start,
 			   position_t end)
 {
-	unsigned rows = 0, bytes;
+	unsigned rows = 0, bytes, max_rows = window->rows + 1;
 
-	for (rows = 0; start < end; rows++, start += bytes)
+	for (rows = 0; start < end && rows < max_rows; rows++, start += bytes)
 		if (!(bytes = find_row_bytes(window->view, start, 0,
 					     window->columns)))
 			break;
@@ -301,15 +328,21 @@ static position_t find_row_start(struct window *window, position_t position,
 
 static void new_start(struct view *view, position_t start)
 {
-	locus_set(view, view->window->start, start ? start-1 : UNSET);
+	if (start)
+		view_char_prior(view, start, &start);
+	else
+		start = UNSET;
+	locus_set(view, view->window->start, start);
 	view->window->last_dirties = ~0;
 }
 
 static position_t screen_start(struct view *view)
 {
 	position_t start = locus_get(view, view->window->start);
-	if (start++ == UNSET)
+	if (start == UNSET)
 		start = 0;
+	else
+		view_char(view, start, &start);
 	return start;
 }
 
@@ -321,17 +354,31 @@ static position_t focus(struct window *window)
 					      find_line_start(view, cursor));
 	position_t start = screen_start(view);
 	position_t end, at;
-	unsigned above, below;
+	unsigned above = 0, below;
 	size_t bytes;
 
-	start = find_row_start(window, start,
-			       find_line_start(view, start));
+	start = find_row_start(window, start, find_line_start(view, start));
+
+	/* Scroll by single lines when cursor is just one row out of view. */
+	if (cursorrow < start &&
+	    start == cursorrow + find_row_bytes(view, cursorrow,
+						0, window->columns)) {
+		start = cursorrow;
+		goto done;
+	}
 	if (cursorrow >= start &&
-	    (above = count_rows(window, start, cursorrow)) < window->rows) {
+	    (above = count_rows(window, start, cursorrow)) == window->rows) {
+		start += find_row_bytes(view, start, 0, window->columns);
+		above--;
+		goto done;
+	}
+
+	if (cursorrow >= start && above < window->rows) {
 		for (below = 1, at = cursorrow;
 		     above + below < window->rows;
 		     below++, at += bytes)
-			if (!(bytes = find_row_bytes(view, at, 0, window->columns)))
+			if (!(bytes = find_row_bytes(view, at, 0,
+						     window->columns)))
 				break;
 		if (above + below == window->rows || !start)
 			goto done;
@@ -397,10 +444,9 @@ static Boolean_t lame_tab(struct view *view, position_t offset)
 	return TRUE;
 }
 
-static unsigned paintch(struct window *window, Unicode_t ch, unsigned row,
-			unsigned column,
-			position_t at, position_t cursor, position_t mark,
-			unsigned *brackets, rgba_t fgrgba)
+static int paintch(struct window *window, Unicode_t ch, int row, int column,
+		   position_t at, position_t cursor, position_t mark,
+		   unsigned *brackets, rgba_t fgrgba)
 {
 	rgba_t bgrgba = window->bgrgba;
 	unsigned tabstop = window->view->text->tabstop;
@@ -413,18 +459,22 @@ static unsigned paintch(struct window *window, Unicode_t ch, unsigned row,
 	    (at >= cursor && at < mark ||
 	     at >= mark && at < cursor)) {
 		bgrgba = window->view->mode->selection_bgrgba;
-		fgrgba = 0xff000000;
+		fgrgba = SELECTION_FGRGBA;
 	} else if (at == cursor) {
-		if (window->view->text->flags & TEXT_RDONLY)
-			fgrgba = 0xff000000;
-		else if (at == mark)
-			fgrgba = 0xff00ff00;
+		rgba_t rgba = DEFAULT_CURSORRGBA;
+		if (mark != UNSET)
+			rgba = SELECTING_RGBA;
+		else if (window->view->text->flags & TEXT_RDONLY)
+			rgba = RDONLY_RGBA;
 		else if (text_is_dirty(window->view->text))
-			fgrgba = 0x00ff0000;
+			rgba = DIRTY_RGBA;
+		if (!display_cursor_color(display, rgba))
+			fgrgba = rgba;
 	}
 
 	if (ch == '\t') {
-		rgba_t bg = lame_tab(window->view, at+1) ? 0xff00ff00 : bgrgba;
+		rgba_t bg = lame_tab(window->view, at+1) ?
+				LAMESPACE_BGRGBA : bgrgba;
 		do {
 			display_put(display, window->row + row,
 				    window->column + column++, ' ', fgrgba, bg);
@@ -434,9 +484,10 @@ static unsigned paintch(struct window *window, Unicode_t ch, unsigned row,
 	}
 
 	if (ch < ' ' || ch == 0x7f || IS_FOLDED(ch)) {
-		bgrgba = 0xff000000;
-		fgrgba = 0xffffff00;
-		display_put(display, window->row + row, window->column + column++,
+		fgrgba = FOLDED_FGRGBA;
+		bgrgba = FOLDED_BGRGBA;
+		display_put(display,
+			    window->row + row, window->column + column++,
 			    IS_FOLDED(ch) ? '<' : '^', fgrgba, bgrgba);
 		if (IS_FOLDED(ch))
 			ch = '>';
@@ -448,14 +499,14 @@ static unsigned paintch(struct window *window, Unicode_t ch, unsigned row,
 		if (bgrgba == window->bgrgba &&
 		    lame_space(window->view, at + 1,
 			       tabstop-1 - column % tabstop))
-			bgrgba = 0xff00ff00;
+			bgrgba = LAMESPACE_BGRGBA;
 	} else if (!IS_UNICODE(ch))
-		ch = ' ', bgrgba = 0xff00ff00;
-	else if (brack)
+		ch = ' ', bgrgba = BADCHAR_BGRGBA;
+	else if (brackets && brack)
 		for (; *brack; brack += 2)
 			if (ch == brack[0] && (*brackets)++ & 1 ||
 			    ch == brack[1] && --*brackets & 1) {
-				fgrgba = 0x0000ff00;
+				fgrgba = BLUE_RGBA;
 				break;
 			}
 
@@ -470,11 +521,8 @@ static Boolean_t needs_repainting(struct window *window)
 	position_t cursor = locus_get(view, CURSOR);
 	position_t mark = locus_get(view, MARK);
 
-	if (mark == UNSET)
-		mark = cursor;
-	return	view->text->dirties != window->last_dirties ||
-		window->last_fgrgba != window->fgrgba ||
-		window->last_bgrgba != window->bgrgba ||
+	return	window->repaint ||
+		view->text->dirties != window->last_dirties ||
 		window->last_cursor != cursor ||
 		window->last_mark != mark ||
 		window->last_mode != view->mode;
@@ -482,9 +530,8 @@ static Boolean_t needs_repainting(struct window *window)
 
 static void repainted(struct window *window, position_t cursor, position_t mark)
 {
+	window->repaint = FALSE;
 	window->last_dirties = window->view->text->dirties;
-	window->last_fgrgba = window->fgrgba;
-	window->last_bgrgba = window->bgrgba;
 	window->last_cursor = cursor;
 	window->last_mark = mark;
 	window->last_mode = window->view->mode;
@@ -492,55 +539,92 @@ static void repainted(struct window *window, position_t cursor, position_t mark)
 
 static void paint(struct window *window)
 {
-	position_t at, next;
-	unsigned row, column;
+	sposition_t at;
+	int row, column;
 	struct view *view = window->view;
 	position_t cursor = locus_get(view, CURSOR);
 	position_t mark = locus_get(view, MARK);
+	sposition_t comment_end = -1;
+	sposition_t string_end = -1;
 	unsigned brackets = 1;
+	Boolean_t keywords = !no_keywords &&
+			     window == active_window &&
+			     view->text->keywords;
 
 	title(window);
 
 	at = focus(window);
+	if (keywords && view->text->comment_start) {
+		sposition_t start = view->text->comment_start(view, at);
+		if (start >= 0)
+			comment_end = view->text->comment_end(view, start);
+	}
+
 	for (row = 0; row < window->rows; row++) {
 
 		position_t limit = at + find_row_bytes(view, at, 0,
 						       window->columns);
 		rgba_t fgrgba = window->fgrgba;
-		Boolean_t keywords = view->text->keywords &&
-				     (fgrgba == DEFAULT_FGRGBA ||
-				      !fgrgba);
 		Boolean_t look_for_keyword = keywords;
+		position_t next;
+		unsigned *brackets_ptr = &brackets;
 
 		for (column = 0; at < limit; at = next) {
+
 			Unicode_t ch = view_char(view, at, &next);
-			if (!is_idch(ch) && ch != '#') {
+
+			if ((ch == '/' || ch == '-' || ch == '{') &&
+			    at > comment_end &&
+			    at > string_end &&
+			    keywords &&
+			    view->text->comment_end)
+				comment_end = view->text->comment_end(view, at);
+			if ((ch == '"' || ch == '\'') &&
+			    at > comment_end &&
+			    at > string_end &&
+			    keywords &&
+			    view->text->string_end)
+				string_end = view->text->string_end(view, at);
+
+			if (at <= comment_end) {
+				fgrgba = COMMENT_FGRGBA;
+				brackets_ptr = NULL;
+			} else if (at <= string_end) {
+				fgrgba = STRING_FGRGBA;
+				brackets_ptr = NULL;
+			} else if (!is_idch(ch) && ch != '#') {
 				fgrgba = window->fgrgba;
 				look_for_keyword = keywords;
 			} else if (look_for_keyword && is_keyword(view, at))
-				fgrgba = 0x0000ff00;
+				fgrgba = KEYWORD_FGRGBA;
 			else
 				look_for_keyword = FALSE;
+
 			column = paintch(window, ch, row, column, at,
-					 cursor, mark, &brackets,
+					 cursor, mark, brackets_ptr,
 					 fgrgba);
+
+			if (at == comment_end || at == string_end) {
+				fgrgba = window->fgrgba;
+				brackets_ptr = &brackets;
+				look_for_keyword = keywords;
+			}
 		}
 
-#if 0
 		display_erase(display, window->row + row,
-			      window->column + column, 1,
-			      window->columns - column);
-#endif
+			      window->column + column,
+			      1, window->columns - column - 1);
 		while (column < window->columns)
 			display_put(display, window->row + row,
 				    window->column + column++,
-				    ' ', window->fgrgba, window->bgrgba);
+				    ' ', DEFAULT_FGRGBA, window->bgrgba);
 	}
 
 	repainted(window, cursor, mark);
 }
 
-void window_hint_deleting(struct window *window, position_t offset, size_t bytes)
+void window_hint_deleting(struct window *window, position_t offset,
+			  size_t bytes)
 {
 	struct view *view = window->view;
 	position_t at = screen_start(view);
@@ -579,10 +663,11 @@ void window_hint_deleting(struct window *window, position_t offset, size_t bytes
 		end_column = 0;
 	}
 	display_delete_lines(display, window->row + row, window->column,
-			     lines, window->rows, window->columns);
+			     lines, window->rows - row, window->columns);
 }
 
-void window_hint_inserted(struct window *window, position_t offset, size_t bytes)
+void window_hint_inserted(struct window *window, position_t offset,
+			  size_t bytes)
 {
 	struct view *view = window->view;
 	position_t at = screen_start(view);
@@ -621,7 +706,7 @@ void window_hint_inserted(struct window *window, position_t offset, size_t bytes
 		end_column = 0;
 	}
 	display_insert_lines(display, window->row + row, window->column,
-			     lines, window->rows, window->columns);
+			     lines, window->rows - row, window->columns);
 }
 
 void window_next(struct view *view)
@@ -643,10 +728,16 @@ struct window *window_recenter(struct view *view)
 	struct window *window = view->window;
 	position_t cursor = locus_get(view, CURSOR);
 	position_t start, end;
-	int row;
+	int row, rows, columns;
 
-	if (!window)
+	if (!display)
+		display = display_init();
+	display_get_geometry(display, &rows, &columns);
+	if (rows != display_rows ||
+	    columns != display_columns ||
+	    !(window = view->window))
 		window = window_raise(view);
+
 	start = find_row_start(window, cursor, find_line_start(view, cursor));
 	for (row = 0;
 	     (end = start) && row < window->rows >> 1;
@@ -658,7 +749,30 @@ struct window *window_recenter(struct view *view)
 	return window;
 }
 
-#define OVERLAP 1
+static int page_overlap(struct window *window)
+{
+	static int overlap_percent;
+	int o;
+
+	if (!overlap_percent) {
+		const char *p = getenv("AOEUI_OVERLAP");
+		overlap_percent = 1;
+		if (p) {
+			overlap_percent = atoi(p);
+			if (overlap_percent < 0)
+				overlap_percent = 1;
+			else if (overlap_percent > 100)
+				overlap_percent = 100;
+		}
+	}
+	o = overlap_percent * window->rows / 100;
+	if (o >= window->rows)
+		o = window->rows - 1;
+	if (o < 1)
+		o = 1;
+	return o;
+}
+
 void window_page_up(struct view *view)
 {
 	struct window *window = view->window;
@@ -668,19 +782,22 @@ void window_page_up(struct view *view)
 	size_t bytes;
 
 	if (start) {
+		int overlap = page_overlap(window);
 		for (row = 0;
-		     (end = start) && row + OVERLAP < window->rows;
+		     (end = start) && row + overlap < window->rows;
 		     row += count_rows(window, start, end))
 			start = find_line_start(view, start-1);
-		while (row-- + OVERLAP > window->rows)
-			start += find_row_bytes(view, start, 0, window->columns);
+		while (row-- + overlap > window->rows)
+			start += find_row_bytes(view, start,
+						0, window->columns);
 		new_start(view, start);
 		for (row = 0; row < window->rows-1; row++, start += bytes)
 			if (!(bytes = find_row_bytes(view, start, 0,
 						     window->columns)))
 				break;
 		display_insert_lines(display, window->row, window->column,
-				     1, window->rows, window->columns);
+				     window->rows - overlap,
+				     window->rows, window->columns);
 	}
 	locus_set(view, CURSOR, start);
 }
@@ -691,11 +808,13 @@ void window_page_down(struct view *view)
 	position_t start = screen_start(view);
 	int row;
 	size_t bytes;
+	int overlap = page_overlap(window);
 
-	for (row = 0; row + OVERLAP < window->rows; row++, start += bytes)
+	for (row = 0; row + overlap < window->rows; row++, start += bytes)
 		if (!(bytes = find_row_bytes(view, start, 0, window->columns)))
 			break;
-	display_delete_lines(display, window->row, window->column, 1,
+	display_delete_lines(display, window->row, window->column,
+			     window->rows - overlap,
 			     window->rows, window->columns);
 	new_start(view, start);
 	locus_set(view, CURSOR, start);
@@ -715,23 +834,24 @@ static void window_colors(void)
 	struct window *window, *w;
 
 	static rgba_t colors[][2] = {
-		{ 0x00000000, 0xffffff00 },
-		{ 0x0000ff00, 0xffff0000 },
-		{ 0x0000ff00, 0x7f7f0000 },
-		{ 0x00ff0000, 0xff00ff00 },
-		{ 0x00ff0000, 0x7f007f00 },
-		{ 0xff000000, 0x007f7f00 },
-		{ 0xff000000, 0x00ffff00 },
+		{ BLACK_RGBA, PALE_RGBA(WHITE_RGBA) },
+		{ BLUE_RGBA, YELLOW_RGBA },
+		{ BLUE_RGBA, PALE_RGBA(YELLOW_RGBA) },
+		{ GREEN_RGBA, MAGENTA_RGBA },
+		{ GREEN_RGBA, PALE_RGBA(MAGENTA_RGBA) },
+		{ RED_RGBA, PALE_RGBA(CYAN_RGBA) },
+		{ RED_RGBA, CYAN_RGBA },
 		{ }
 	};
 
-	for (window = window_list; window; window = window->next)
-		window->bgrgba = 0;
-	j = active_window != window_list || active_window->next;
-	active_window->fgrgba = DEFAULT_FGRGBA;
-	active_window->bgrgba = DEFAULT_BGRGBA;
+	if (active_window) {
+		if (active_window->fgrgba != DEFAULT_FGRGBA)
+			active_window->repaint = TRUE;
+		active_window->fgrgba = DEFAULT_FGRGBA;
+		active_window->bgrgba = DEFAULT_BGRGBA;
+	}
 	for (window = window_list; window; window = window->next) {
-		if (window->bgrgba)
+		if (window == active_window)
 			continue;
 		for (j = 0; colors[j][1]; j++) {
 			for (w = window_list; w; w = w->next)
@@ -742,8 +862,11 @@ static void window_colors(void)
 			if (!w)
 				break;
 		}
-		window->fgrgba = colors[j][0];
-		window->bgrgba = colors[j][1];
+		if (window->bgrgba != colors[j][1]) {
+			window->repaint = TRUE;
+			window->fgrgba = colors[j][0];
+			window->bgrgba = colors[j][1];
+		}
 	}
 }
 
